/*

Copyright (C) 2009 Jaroslav Barton <djaara@djaara.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include <QDBusConnectionInterface>
#include <QDBusInterface>
#include <QDebug>

#include "FMDbusConn.h"
#include "FingerManagerDevicesModel.h"

FMDbusConn *conn = NULL;
int		ocount = 0; // number of opened connection

/**
 * Hides constructor
 */
FMDbusConn* FMDbusConn::getConn() {
	if (ocount == 0) {
		conn = new FMDbusConn();
	}

	ocount++;
	return(conn);
}

/**
 * Hides destructor
 */
void FMDbusConn::releaseConn() {
	if (ocount > 0) {
		ocount--;
	}

	if (ocount == 0){
		delete this;
	}
}

/**
 * FMDbusConn constructor
 */
FMDbusConn::FMDbusConn() 
    : m_managerIface(NULL), 
      m_claimedDevice(NULL), 
      m_isEnrollStopped(true),
	  m_releaseInProgress(false) 
{
	if (!QDBusConnection::systemBus().isConnected()) {
		qDebug("Cannot connect to D-Bus system bus.");
	}
}

/**
 * FMDbusConn destructor
 */
FMDbusConn::~FMDbusConn() {
	delete m_managerIface;
}

/**
 * Start D-Bus daemon (requires some retries)
 * @param retry maximum retries while waiting for fprintd D-Bus service start
 */
QDBusInterface* FMDbusConn::getManagerIface(int retry) {
	if (!retry) {
		return NULL;
	}
	
	if (m_managerIface && m_managerIface->isValid()) {
		return m_managerIface;
	}
	
	// Connect to Fprint manager on system bus
	QDBusConnection bus = QDBusConnection::systemBus();
    QDBusInterface *iface = NULL;

    for (int i = 0; i < retry; ++i) {
    	 iface = new QDBusInterface("net.reactivated.Fprint",
				                    "/net/reactivated/Fprint/Manager",
                                    "net.reactivated.Fprint.Manager",
				                    bus, this);
	    // Wait while D-Bus Fprintd daemon starts
	    if (iface && iface->isValid()) {
		    break;
    	}
    }

	return iface;
}

/**
 * Get device communication interface
 * @param devicePath device which interface is required
 * @return QDBusInterface
 */
QDBusInterface* FMDbusConn::getDeviceInterface(const QString &devicePath) {
	QDBusInterface *iface = new QDBusInterface( "net.reactivated.Fprint",
		                                    devicePath,
                                        	"net.reactivated.Fprint.Device",
                                    		QDBusConnection::systemBus());
	if (iface && iface->isValid()) {
    	return iface;
    } else {
        return NULL;
    }
}

/**
 * Returns name associated with device
 * @param deviceiPath device which name is required
 * @return QString
 */
QString FMDbusConn::getDeviceName(const QString &devicePath) {
	QString deviceName;
	QDBusInterface *iface = getDeviceInterface(devicePath);
	if (iface && iface->isValid()) {
		deviceName = iface->property("name").toString();
	} else {
		deviceName = devicePath;
	}
    
    delete iface;
	return deviceName;
}

/**
 * getDevices - get list of installed devices
 * @return FingerManagerDevicesModel
 */
FingerManagerDevicesModel* FMDbusConn::getDevices() {
	QDBusInterface *iface = getManagerIface();
	FingerManagerDevicesModel *model = new FingerManagerDevicesModel();
	
	if (iface && iface->isValid()) {
		qDebug("D-Bus interface acquired! :-)");
		QDBusReply<QList<QDBusObjectPath> > devices = iface->call("GetDevices");
		if (devices.isValid()) {
			foreach (QDBusObjectPath device, devices.value()) {
				model->addDevice(device.path(), getDeviceName(device.path()));
qDebug() << __FILE__ << __LINE__ << __func__ << device.path();  
qDebug() << __FILE__ << __LINE__ << __func__ << getDeviceName(device.path());  
			}
		}

		if (model->rowCount(QModelIndex()) > 0) { // found at least one device
			QDBusReply<QDBusObjectPath> defaultDevice = iface->call("GetDefaultDevice");
			if (defaultDevice.isValid()) {
				model->setDefaultDevice(defaultDevice.value().path());
			}
		}
	} else {
		qDebug("D-Bus interface cannot be acquired! :-(");
	}

	return(model);
}

/**
 * getFingers - get list of enrolled fingerprints for user on selected device
 * @param devicePath device
 * @param username username
 * @return QStringList
 */
QStringList FMDbusConn::getFingers(const QString &devicePath, const QString &username) {
	QDBusInterface *iface = getDeviceInterface(devicePath);
	QStringList result;

	if (iface && iface->isValid()) {
		QDBusReply<QStringList> r = iface->call("ListEnrolledFingers", username);
		if (r.isValid()) {
			result = r.value();
		} else {
			QDBusError e = iface->lastError();
			qDebug() << e.message();
		}
	}

	delete(iface);
	return(result);
}

/**
 * deleteAllEnrolledFingers - delete all enrolled fingers for user on selected device
 * @param devicePath device
 * @param username username
 * @return bool
 */
bool FMDbusConn::deleteAllEnrolledFingers(const QString &devicePath, const QString &username) {
	QDBusInterface *iface = getDeviceInterface(devicePath);
	bool retval = false;

	if (iface && iface->isValid()) {
        QDBusMessage message = iface->call("DeleteEnrolledFingers", username);
        if (message.type() == QDBusMessage::ReplyMessage) { 
		    retval = true;
        }
	}

	delete iface;
	return retval;
}

/*******************************************************************************
 * Device informations
 */

/**
 * getEnrollStages - returns number of enroll stages for selected device
 * @return int
 */
int FMDbusConn::getEnrollStages() {
	int stages = 3;
	if (isDeviceClaimed()) {
		// QDBusInterface cannot access properties with dashes in name,
		// so access must be done in other way
		QDBusInterface *iface = new QDBusInterface( "net.reactivated.Fprint",
				                                    m_claimedDevice->path(),
                                    				"org.freedesktop.DBus.Properties",
                                    				QDBusConnection::systemBus());
		if (iface && iface->isValid()) {
			QDBusReply<QVariant> r = iface->call("Get",
					                            "net.reactivated.Fprint.Device",
                            					"num-enroll-stages");
			int s = r.value().toInt();
			if (s > 0) {
				stages = s;
			}
		}
	    
        delete iface;
	}

	return(stages);
}

/**
 * getScanType - return sensor type for selected device
 * @return QString
 */
QString FMDbusConn::getScanType() {
	QString scanType("swipe");

	if (isDeviceClaimed()) {
		QDBusInterface *iface = new QDBusInterface( "net.reactivated.Fprint",
				                                    m_claimedDevice->path(),
                                    				"org.freedesktop.DBus.Properties",
                                    				QDBusConnection::systemBus());
		if (iface && iface->isValid()) {
		    QDBusReply<QVariant> r = iface->call("Get",
					                            "net.reactivated.Fprint.Device",
                                				"scan-type");
			QString t = r.value().toString();

			if (t.length() > 0) {
				scanType = t;
			}
		}

		delete iface;
	}

	return scanType;
}

/*******************************************************************************
 * User enrollment
 */
/**
 * claimDevice - claim device for user
 * @param devicePath device
 * @param username username
 */
void FMDbusConn::claimDevice(const QString &devicePath, const QString &username) {
	if (!m_claimedDevice) {
		QDBusInterface *iface = getDeviceInterface(devicePath);
		if (iface && iface->isValid()) {
			QDBusMessage m = iface->call("Claim", username);
			if (m.type() == QDBusMessage::ErrorMessage) {
				delete(iface);
				m_claimedDevice = NULL;
			} else {
				m_claimedDevice = iface;
				QObject::connect(iface, SIGNAL(EnrollStatus(QString,bool)), this, SLOT(enrollStatusSlot(QString,bool)));
			}
		} else {
			qDebug() << "Error:"<< iface->lastError().message();
			m_claimedDevice = NULL;
			delete(iface);
		}
	}
}

/**
 * releaseDevice - release claimed device
 */
void FMDbusConn::releaseDevice() {
	if (m_claimedDevice && !m_releaseInProgress) {
		m_releaseInProgress = true;
		// no locking on shared variable m_releaseInProgress, should be ok
		if (m_claimedDevice->isValid()) {
			if (!m_isEnrollStopped) {
                enrollStop();
            }

			QDBusMessage m = m_claimedDevice->call("Release");
			if (m.type() == QDBusMessage::ErrorMessage) {
				qDebug() << "  Release device:" << m.errorMessage();
			} else {
                m_releaseInProgress = false;
            }
		}

		delete m_claimedDevice;
		m_claimedDevice = NULL;
	}
}

/**
 * isDeviceClaimed
 * @return bool true if is device claimed, otherwise false
 */
bool FMDbusConn::isDeviceClaimed() {
	if (m_claimedDevice && m_claimedDevice->isValid()) {
		return true;
	} else {
		return false;
	}
}

/**
 * enrollStatusSlot- forward message from fprintd D-Bus service
 * @param status enroll status
 * @param result enroll result 
 */
void FMDbusConn::enrollStatusSlot(QString status, bool result) {
	emit EnrollStatus(status, result);
}

/**
 * enrollStart - start enrollment for selected finger
 * @param finger selected finger
 */
void FMDbusConn::enrollStart(const QString &fingerName) {
	if (isDeviceClaimed()) {
		if (m_isEnrollStopped) {
			m_isEnrollStopped = false;
			m_claimedDevice->call("EnrollStart", fingerName);
		}
	}
}

/**
 * enrollStop - stop enrollment
 */
void FMDbusConn::enrollStop() {
	if (isDeviceClaimed()) {
		if (!m_isEnrollStopped) {
			m_isEnrollStopped = true;
			m_claimedDevice->call("EnrollStop");
		}
	}
}
