/*

Copyright (C) 2009 Jaroslav Barton <djaara@djaara.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include <KConfig>
#include <KConfigGroup>
#include <KLocale>
#include <KSeparator>
#include <QWidget>
#include <QPixmap>
#include <QButtonGroup>
#include <QSize>
#include <QDebug>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <KPushButton>
#include <KMessageBox>
#include <KPluginFactory>
#include <KPluginLoader>

#include "FingerManagerWindow.h"
#include "FingerManagerDevicesModel.h"
#include "FMDbusConn.h"
#include "FingerNames.h"
#include "FMEnroll.h"

int	y_base;

K_PLUGIN_FACTORY(FingerManagerWindowFactory,
	registerPlugin<FingerManagerWindow>();
	)

K_EXPORT_PLUGIN(FingerManagerWindowFactory("kcmkfingermanager"))

const static char fingerManagerVersion[] = "v0.0.5";

/**
 * FingerManagerWindow contructor
 * @param parent parent widget
 */
FingerManagerWindow::FingerManagerWindow(QWidget *parent, const QVariantList &) 
    : KCModule(FingerManagerWindowFactory::componentData(), parent) 
{
	m_aboutData = new KAboutData(   "kfingermanager", 
                                    "KFingerManager",
		                            ki18n("Manage Enrolled Fingers"), 
                                    fingerManagerVersion,
                    		        ki18n("Manage Enrolled Fingers of User"), 
                                    KAboutData::License_GPL,
                    		        ki18n("(c) 2009 Jaroslav Barton"), 
                                    ki18n(0L));
	m_aboutData->addAuthor(ki18n("Jaroslav Barton"), ki18n("Maintainer"), "djaara@djaara.net");
	setAboutData(m_aboutData);
	setButtons(KCModule::NoAdditionalButton);

	KUser user;
	m_loginName = user.loginName();

	// Prepare component
	initComponents();
	loadConfig(m_config);
	retranslate();

	// Handle all check buttons
	m_btng->setExclusive(false);
	connect(m_btng, SIGNAL(buttonPressed(int)), SLOT(fingerSelectedSlot(int)));

	connect(m_devices, SIGNAL(currentIndexChanged(int)), this, SLOT(deviceChangedSlot(int)));
	connect(m_deleteAllBtn, SIGNAL(pressed()), this, SLOT(deleteAllSlot()));

	// Devices model for combo box
	FingerManagerDevicesModel* model = m_dbc->getDevices();
	m_devices->setModel(model);
	m_devices->setCurrentIndex(model->getDefaultDeviceIndex());

	// Layout
	QVBoxLayout *mainPanelLayout = new QVBoxLayout(m_box);
	
    QHBoxLayout *handsLayout = new QHBoxLayout();
	handsLayout->addStretch();
	handsLayout->addWidget(m_handsPanel);
	handsLayout->addStretch();

	QHBoxLayout *devicesLayout = new QHBoxLayout();
	devicesLayout->setSpacing(KDialog::spacingHint());
	devicesLayout->setMargin(0);
	devicesLayout->addWidget(m_deviceLabel);
	devicesLayout->addWidget(m_devices);
	devicesLayout->addStretch();
	devicesLayout->addWidget(m_deleteAllBtn);

	mainPanelLayout->addLayout(handsLayout);
	mainPanelLayout->addWidget(m_separator);
	mainPanelLayout->addLayout(devicesLayout);

	QVBoxLayout *mainLayout = new QVBoxLayout(this);
	mainLayout->setSpacing(KDialog::spacingHint());
	mainLayout->setMargin(0);
	mainLayout->addWidget(m_box);
	mainLayout->addStretch();

	setMinimumSize(m_box->minimumSizeHint());
}

/**
 * Initialize main components of FingerManagerWindow
 */
void FingerManagerWindow::initComponents() {
	m_dbc = FMDbusConn::getConn();

	// config file
	m_config = new KConfig("kfingerrc");

	// box around UI
	m_box = new QGroupBox(this);

	// pane with hands image
	//and image
	m_handsPanel = new QWidget(this);
	m_handsLabel = new QLabel(m_handsPanel);
	// and with buttons
	m_btng = new QButtonGroup(m_handsPanel);

	// m_separator
	m_separator = new KSeparator(m_box);

	// device section
	m_deviceLabel = new QLabel(this);
	m_devices = new QComboBox(this);

	m_deleteAllBtn = new KPushButton(this);
	m_deleteAllBtn->setEnabled(false);

	for (int i = 0; i < MAX_FINGERS; i++) {
		m_fingers[i] = NULL;
	}
}

/**
 * FingerManagerWindow destructor
 */
FingerManagerWindow::~FingerManagerWindow() {
	for (int i = 0; i < MAX_FINGERS; i++) {
		if (m_fingers[i]) {
            delete m_fingers[i];
        }
	}

	if (m_dbc) {
        m_dbc->releaseConn();
    }

	delete m_config;
}
/**
 * Handle checkbox status changes
 * @param idx selected finger index
 */
void FingerManagerWindow::fingerSelectedSlot(int idx) {
	int devid = m_devices->currentIndex();
	if (devid > -1) {
		if (m_fingers[idx]->isChecked()) {
			KMessageBox::sorry(this, i18n("Deletion of single "
				"fingerprint is not supported in fprintd D-Bus "
				"daemon"));
			m_fingers[idx]->setChecked(true);
		} else {
			FMEnroll *e = new FMEnroll(
				                ((FingerManagerDevicesModel*)m_devices->model())->getDeviceModel(devid),
				                m_loginName,
				                idx,
				                this);
			int retval = e->exec();
			delete(e);
			if (retval == QDialog::Rejected) {
				m_fingers[idx]->setChecked(false);
			} else {
                m_fingers[idx]->setChecked(true);
            }
		}
	}
}

/**
 * Retrieves enrolled fingers from dbus and set checked corresponding checkboxes
 * @param fngrs list of enrolled fingers
 */
void FingerManagerWindow::processFingers(const QStringList &fngrs) {
	for (int i = 0; i < 10; i++) {
		if (m_fingers[i]) {
            m_fingers[i]->setChecked(false);
        }
	}

	for (int i = 0; i < fngrs.size(); i++) {
		QString finger = fngrs.at(i);
		for (int j = 0; j < MAX_FINGERS; j++) {
			if (finger == fingerNames[j].iface) {
				if (m_fingers[j]) {
                    m_fingers[j]->setChecked(true);
                }
			}
		}
	}
}

void FingerManagerWindow::enableAllFingers(bool enabled) {
	for (int i = 0; i < MAX_FINGERS; i++) {
		if (m_fingers[i]) {
            m_fingers[i]->setEnabled(enabled);
        }
	}
}

/**
 * Hook after search enrolled fingerprints for selected device
 * @param idx new selected device
 */
void FingerManagerWindow::deviceChangedSlot(int idx) {
	if (idx > -1) {
		QStringList fingers = m_dbc->getFingers(((FingerManagerDevicesModel*)m_devices->model())->getDevicePath(idx), m_loginName);
		processFingers(fingers);
		m_deleteAllBtn->setEnabled(true);
		enableAllFingers(true);
	} else {
		m_deleteAllBtn->setEnabled(false);
		enableAllFingers(false);
	}
}

/**
 * Verify deletion of enrolled fingerprints on selected device for current user
 */
void FingerManagerWindow::deleteAllSlot() {
	int idx = m_devices->currentIndex();
	if (idx > -1) {
		// Warning message
		QString device = ((FingerManagerDevicesModel*)m_devices->model())->
				getDeviceName(idx);
		QString msg = i18n("Do you really want delete all fingerprints "
					"on device \"%1\" for user \"%2\"", device, m_loginName);
		// Show Yes/No message box
		int result = KMessageBox::questionYesNo(this, msg);
		if (result == KMessageBox::Yes) {
			if (m_dbc->deleteAllEnrolledFingers(((FingerManagerDevicesModel*)m_devices->model())->getDevicePath(idx),m_loginName)) {
				deviceChangedSlot(idx);
			};
		}
	}
}

/**
 * createCheckBox - create checkbox on selected coordinates with finger index
 * @param idx finger index
 * @param x x coordinate
 * @param y y coordinate
 * @return new checkbox
 */
QCheckBox* FingerManagerWindow::createCheckBox(int idx, int x, int y) {
	QCheckBox *cb = new QCheckBox(m_handsPanel);
	QSize size = cb->sizeHint();
	cb->setGeometry(QRect(x - (size.width()/2), 
			        y_base + y - (size.height()/2),
			        size.height(),
			        size.height()));
	m_btng->addButton(cb, idx);

	return(cb);
}

/**
 * loadConfig - config file loader
 * @param settings configuration settings
 */
void FingerManagerWindow::loadConfig(KConfig *config) {
	// Config helper
	KConfigGroup cfgGroup;
	
	// Load image
	cfgGroup = config->group("image");
	y_base = cfgGroup.readEntry("base", 20);
	
	QString filename = cfgGroup.readEntry("filename", "hands.png");
	QPixmap image = QPixmap(filename);
 	if (!image.isNull()) {
		m_handsLabel->setPixmap(image);
	} else {
		m_handsLabel->setText(i18n("File not found: %1",filename));
	}
	// set label position
	m_handsLabel->setGeometry(QRect(0, y_base, image.width(), image.height()));

	// Set m_handsPanel dimension
	m_handsPanel->setMinimumWidth(m_handsLabel->width());
	m_handsPanel->setMinimumHeight(m_handsLabel->height());

	// Load checkboxes
	cfgGroup = config->group("fingers");
	for (int i = 0; i < MAX_FINGERS; i++) {
		QString name;
		
		name.sprintf("%d/x", i+1);
		int x = cfgGroup.readEntry(name, 0);
		
		name.sprintf("%d/y", i+1);
		int y = cfgGroup.readEntry(name, 0);
		
		if (x && y) {
			m_fingers[i] = createCheckBox(i, x, y);
		} else {
			m_fingers[i] = NULL;
		}
	}

	enableAllFingers(false);
}

/**
 * Actualize all texts in window after language changes
 */
void FingerManagerWindow::retranslate() {
	// Window title
	setWindowTitle(i18n("Finger manager..."));

	// What's this
	setWhatsThis(i18n("Finger Manager Application"));

	m_deviceLabel->setText(i18n("Select device:"));
	m_deleteAllBtn->setText(i18n("Delete all"));
	m_deleteAllBtn->setToolTip(i18n("Delete all enrolled fingerprints on selected device."));

	m_box->setTitle(i18n("Select finger to enroll or modify"));

	for (int i = 0; i < MAX_FINGERS; i++) {
		if (m_fingers[i]) {
			m_fingers[i]->setWhatsThis(i18n(fingerNames[i].name));
			m_fingers[i]->setToolTip(i18n(fingerNames[i].name));
		}
	}
}

